/* raw2png.c
 *
 * Copyright (C) 2006, Joel Andersson <bja@kth.se>.
 * Copyright (C) 2002, Gunnar Monell <gmo@linux.nu>.
 * Copyright (C) 2000, Stephan Beyer <fbshot@s-beyer.de>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>  
#include <stdlib.h> 
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <getopt.h>
#include <png.h>
#include <sys/mman.h>
#include <getopt.h>

#define MAX_LEN 512

static void usage()
{
  printf("Usage: raw2png [OPTIONS] FILE FILE\n\n");
  printf("Where options include:\n");
  printf("  -w width      width of raw image\n");
  printf("  -h height     height of raw image\n");
  printf("  -d depth      bits per pixel of raw image\n");
}


static void convert1555to32(int width, int height, 
			    unsigned char *inbuffer, 
			    unsigned char *outbuffer)
{
  unsigned int i;

  for (i=0; i < (unsigned int) height*width*2; i+=2)
    {
      /* BLUE  = 0 */
      outbuffer[(i<<1)+0] = (inbuffer[i+1] & 0x7C) << 1;
      /* GREEN = 1 */
      outbuffer[(i<<1)+1] = (((inbuffer[i+1] & 0x3) << 3) | 
			     ((inbuffer[i] & 0xE0) >> 5)) << 3;
      /* RED   = 2 */
      outbuffer[(i<<1)+2] = (inbuffer[i] & 0x1f) << 3;
      /* ALPHA = 3 */
      outbuffer[(i<<1)+3] = '\0'; 
    }
}

static void convert565to32(int width, int height, 
			   unsigned char *inbuffer, 
			   unsigned char *outbuffer)
{
  unsigned int i;

  for (i=0; i < (unsigned int) height*width*2; i+=2)
    {
      /* BLUE  = 0 */
      outbuffer[(i<<1)+0] = (inbuffer[i] & 0x1f) << 3;
      /* GREEN = 1 */
      outbuffer[(i<<1)+1] = (((inbuffer[i+1] & 0x7) << 3) | 
			     (inbuffer[i] & 0xE0) >> 5) << 2;	
      /* RED   = 2 */
      outbuffer[(i<<1)+2] = (inbuffer[i+1] & 0xF8);
      /* ALPHA = 3 */
      outbuffer[(i<<1)+3] = '\0'; 
    }
}

static void convert888to32(int width, int height, 
			   unsigned char *inbuffer, 
			   unsigned char *outbuffer)
{
  unsigned int i;

  for (i=0; i < (unsigned int) height*width; i++)
    {
      /* BLUE  = 0 */
      outbuffer[(i<<2)+0] = inbuffer[i*3+0];
      /* GREEN = 1 */
      outbuffer[(i<<2)+1] = inbuffer[i*3+1];
      /* RED   = 2 */
      outbuffer[(i<<2)+2] = inbuffer[i*3+2];
      /* ALPHA */
      outbuffer[(i<<2)+3] = '\0';
    }
}

static void write_PNG(unsigned char *outbuffer, const char *filename, 
		      int width, int height, int interlace)
{
  int i;
  int bit_depth=0, color_type;

  png_bytep row_pointers[height];
  png_structp png_ptr;
  png_infop info_ptr;

  FILE *outfile = fopen(filename, "wb");

  for (i=0; i<height; i++) {
    row_pointers[i] = outbuffer + i * 4 * width;
  }
    
  if (!outfile) {
    fprintf (stderr, "Error: Couldn't fopen %s.\n", filename);
    exit(EXIT_FAILURE);
  }
    
  png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, 
				    (png_voidp) NULL, (png_error_ptr) NULL, (png_error_ptr) NULL);
    
  if (!png_ptr) {
    fprintf(stderr, "Error: Couldn't create PNG write struct.");
    exit(EXIT_FAILURE);
  }
    
  info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr) {
    png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
    fprintf(stderr, "Error: Couldn't create PNG info struct.");
  }
    
  png_init_io(png_ptr, outfile);
    
  png_set_compression_level(png_ptr, Z_BEST_COMPRESSION);
    
  bit_depth = 8;
  color_type = PNG_COLOR_TYPE_RGB_ALPHA;
  png_set_invert_alpha(png_ptr);
  png_set_bgr(png_ptr);

  png_set_IHDR(png_ptr, info_ptr, width, height, 
	       bit_depth, color_type, interlace, 
	       PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
    
  png_write_info(png_ptr, info_ptr);
    
  png_write_image(png_ptr, row_pointers);
    
  png_write_end(png_ptr, info_ptr);

  png_destroy_write_struct(&png_ptr, &info_ptr);
    
  if (outfile != NULL) {
    fclose(outfile);
  }
}
	

static void convert_and_write(unsigned char *inbuffer, const char *filename, 
			      int width, int height, int bits, int interlace)
{
  size_t bufsize = (size_t) width * height * 4;

  unsigned char *outbuffer = malloc(bufsize);

  if (outbuffer == NULL) {
    fprintf(stderr, "Not enough memory");
  }
    
  memset(outbuffer, 0, bufsize);

  switch(bits) 
    {

    case 15:
      convert1555to32(width, height, inbuffer, outbuffer);
      write_PNG(outbuffer, filename, width, height, interlace);
      break;

    case 16:
      convert565to32(width, height, inbuffer, outbuffer);
      write_PNG(outbuffer, filename, width, height, interlace);
      break;

    case 24:
      convert888to32(width, height, inbuffer, outbuffer);
      write_PNG(outbuffer, filename, width, height, interlace);
      break;

    case 32:
      /* No conversion needed */
      write_PNG(inbuffer, filename, width, height, interlace);
      break;

    default:
      fprintf(stderr, "%d bits per pixel are not supported! ", bits);
      exit(EXIT_FAILURE);
    }
        
  (void) free(outbuffer);
}


int main(int argc, char **argv)
{
  int optc;
  int fd;

  const char *infile;
  const char *outfile;
  unsigned char *buf;

  int width = 0;
  int height = 0;
  int depth = 0;

  for (;;) {

    optc = getopt(argc, argv, "w:h:d:");

    if (optc == -1) {
      break;
    }

    switch (optc) {

    case 'd':
      depth = atoi(optarg);
      break;

    case 'h':
      height = atoi(optarg);
      break;

    case 'w':
      width = atoi(optarg);
      break;    

    default:
      usage(argv[0]);
    }
  }
    
  if ((optind == argc) || (argc - optind != 2)) {
    usage(argv[0]);
    return 1;
  }

  infile = argv[optind];
  outfile = argv[optind+1];

  fd = open(infile, O_RDONLY);
  if (fd < 0) {
    perror("raw2png: open");
    return 0;
  }

  int mm_prot = PROT_READ;
  int mm_flags = MAP_PRIVATE;
  int mm_size = width * height * depth;

  buf = mmap(0, mm_size, mm_prot, mm_flags, fd, 0);
  if (buf == MAP_FAILED) {
    perror("raw2png: mmap");
    close(fd);
    return -1;
  }

  convert_and_write(buf, outfile, width, height, depth, 0);
   
  munmap(buf, mm_size);

  close(fd);

  return 0;
}

